from __future__ import division, print_function, unicode_literals

import random
import types
import unittest

try:
    import unittest.mock as mock
except ImportError:
    import mock
from pyoma.browser.db import *
from pyoma.browser import tablefmt
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio import SeqIO


class TestHelperFunctions(unittest.TestCase):
    def test_counter(self):
        self.assertEqual(0, count_elements([]))
        self.assertEqual(3, count_elements("abc"))
        recarray = numpy.zeros(2, dtype=[("A", "i4"), ("B", "f8")])
        self.assertEqual(2, count_elements(recarray))


def find_path_to_test_db(dbfn="TestDb.h5"):
    """We try to load the dbfn first from the same directory, and afterwards from
    the path given by the PYOMA_DB_PATH environment variable.

    :returns: path to database
    :rtype: str
    :raises IOError: if db does not exist."""

    def is_hdf5_file(fname):
        with open(fname, "rb") as fh:
            return fh.read(4) == b"\x89HDF"

    path1 = os.path.join(os.path.dirname(__file__), dbfn)
    if os.path.isfile(path1) and is_hdf5_file(path1):
        return path1
    path2 = os.path.abspath(os.path.join(os.getenv("PYOMA_DB_PATH", "./"), dbfn))
    if os.path.isfile(path2) and is_hdf5_file(path2):
        return path2
    else:
        raise IOError("cannot access {}. (Tried {} and {})".format(dbfn, path1, path2))


class DatabaseTests(unittest.TestCase):
    db = None

    @classmethod
    def setUpClass(cls):
        path = find_path_to_test_db("TestDb.h5")
        logger.info("Loading {} for DatabaseTests".format(path))
        cls.db = Database(path)

    @classmethod
    def tearDownClass(cls):
        cls.db.get_hdf5_handle().close()

    def test_get_vpairs_of_entry_with_orthologs(self):
        for entry_nr, exp_vps_cnt in [(12, 3), (1, 0), (4, 1)]:
            vps = self.db.get_vpairs(entry_nr)
            self.assertTrue(isinstance(vps, numpy.ndarray))
            self.assertEqual(exp_vps_cnt, len(vps))
            self.assertEqual(
                sorted(["EntryNr1", "EntryNr2", "RelType", "Distance", "Score"]),
                sorted(vps.dtype.fields.keys()),
            )

    def test_neighborhood_close_to_boundary(self):
        query, window = 3, 7
        neighbors, idx = self.db.neighbour_genes(query, window)
        self.assertEqual(query - 1, idx)
        self.assertEqual(neighbors["EntryNr"][idx], query)
        expected_entry_nrs = numpy.arange(1, query + window + 1, dtype="i4")
        self.assertTrue(numpy.array_equal(expected_entry_nrs, neighbors["EntryNr"]))

    def test_hog_family(self):
        entry = numpy.zeros(1, dtype=tables.dtype_from_descr(tablefmt.ProteinTable))
        entry["OmaHOG"] = b""
        with self.assertRaises(Singleton):
            self.db.hog_family(entry[0])
        entry["OmaHOG"] = b"HOG:0000523"
        self.assertEqual(523, self.db.hog_family(entry[0]))

    def test_orthoxml(self):
        xml = self.db.get_orthoxml(33).decode()
        # we simply check that orthoxml starts with <?xml and ends with an orthoxml tag
        self.assertTrue(xml.startswith("<?xml "))
        self.assertTrue(xml.endswith("</orthoXML>\n"))

    def test_hog_lex_range(self):
        cases = [
            (b"HOG:001", (b"HOG:001", b"HOG:002")),
            (b"HOG:001.1a.2b", (b"HOG:001.1a.2b", b"HOG:001.1a.2c")),
        ]
        for hog, rng in cases:
            self.assertEqual(rng, self.db._hog_lex_range(hog))

    def test_fam_member(self):
        memb = self.db.member_of_fam(1)
        self.assertEqual(2, len(memb))

    def test_hog_members(self):
        cases = [("Eukaryota", 5), ("Fungi", 3), ("Taphrinomycotina", 0)]
        for level, exp_member_cnt in cases:
            if exp_member_cnt == 0:
                with self.assertRaises(ValueError):
                    self.db.hog_members(12, level)
            else:
                self.assertEqual(len(self.db.hog_members(12, level)), exp_member_cnt)

    def test_hogids_at_level(self):
        cases = [
            [(2, "Ascomycota"), numpy.array([b"HOG:0000002"])],
            [
                (2, "Saccharomyces cerevisiae (strain ATCC 204508 / S288c)"),
                numpy.array([b"HOG:0000002.2a", b"HOG:0000002.2b"]),
            ],
        ]

        for case in cases:
            args, expected = case
            levels = self.db.get_subhogids_at_level(*args)
            self.assertTrue(
                numpy.array_equal(expected, levels),
                "test of tes_hogids_at_level failed for {}: {}".format(args, levels),
            )

    def test_get_hog_without_level(self):
        for case in ("HOG:0000002.2b", "HOG:0000001"):
            with self.subTest(case=case):
                hog = self.db.get_hog(case)
                self.assertTrue(hog["IsRoot"])
                self.assertEqual(case, hog["ID"].decode())

    def test_get_parent_hogs(self):
        for hog_id, level in (
            ("HOG:0000474.1d", "Ascomycota"),
            ("HOG:0000165", "Fungi"),
        ):
            with self.subTest(hog_id=hog_id, level=level):
                hogs = self.db.get_parent_hogs(hog_id, level)
                try:
                    k = hog_id.index(".")
                    base_id = hog_id[:k]
                except ValueError:
                    base_id = hog_id
                self.assertEqual(base_id, hogs[0].hog_id)
                self.assertTrue(hogs[0].is_root)
                self.assertEqual(hog_id, hogs[-1].hog_id)
                self.assertEqual(level, hogs[-1].level)

    def test_get_subhogs(self):
        for hog_id, level in (
            ("HOG:0000474.1d", "Ascomycota"),
            ("HOG:0000165", "Fungi"),
        ):
            with self.subTest(hog_id=hog_id, level=level):
                hogs = list(self.db.get_subhogs(hog_id, level))
                self.assertEqual(hogs[0].hog_id, hog_id)
                self.assertNotIn(level, [h.level for h in hogs])

    def test_get_subhogs_with_subids(self):
        for hog_id, level, exp_subhogid in (
            ("HOG:0000474", "Eukaryota", "HOG:0000474.1d.2c"),
            ("HOG:0000165", "Fungi", "HOG:0000165.1a"),
        ):
            with self.subTest(hog_id=hog_id, level=level):
                hogs = list(
                    self.db.get_subhogs(hog_id, level=level, include_subids=True)
                )
                self.assertIn(exp_subhogid, [h.hog_id for h in hogs])
                self.assertNotIn(level, [h.level for h in hogs])

    def test_get_hog_with_level(self):
        for hog_id, level in (
            ("HOG:0000002.2b", "Saccharomyces cerevisiae (strain ATCC 204508 / S288c)"),
            ("HOG:0000005.1b", "Saccharomycetaceae"),
            ("HOG:0000005", "Fungi"),
        ):
            with self.subTest(hog_id=hog_id, level=level):
                hog = self.db.get_hog(hog_id, level=level)
                self.assertEqual(hog_id, hog["ID"].decode())
                self.assertEqual(level, hog["Level"].decode())

    def test_member_of_hog_id(self):
        cases = [
            [("HOG:0000082.1b", None), 2],
            [("HOG:0000082.1b", "Saccharomycetaceae"), 2],
            [
                (
                    "HOG:0000082.1b",
                    "Saccharomyces cerevisiae (strain ATCC 204508 / S288c)",
                ),
                1,
            ],
            [("HOG:0000082.1a", "Mammalia"), 0],
        ]
        for args, expected_len in cases:
            members = self.db.member_of_hog_id(*args)
            self.assertEqual(len(members), expected_len)

    def test_sorted_genomes(self):
        for root in ("YEAST", "ASHGO"):
            order = self.db.id_mapper["OMA"].species_ordering(root)
            self.assertEqual(
                order[root],
                0,
                "{} should be first genome, but comes at {}".format(root, order[root]),
            )

    def test_main_isoform_from_species_id(self):
        query = "PLAF7"
        rng = self.db.id_mapper["OMA"].genome_range(query)
        mains = self.db.main_isoforms(query)
        self.assertGreaterEqual(rng[1] - rng[0] + 1, len(mains))

    def test_valid_protein_seq_check(self):
        for c in (b"ADRIAN", "adrian", b"XaMGAt"):
            self.assertTrue(
                self.db.seq_search.contains_only_valid_chars(c),
                "'{}' should be a valid protein seq".format(c),
            )
        for c in (b"XBra", "T5HSE", "xxnnebqpynn", " "):
            self.assertFalse(
                self.db.seq_search.contains_only_valid_chars(c),
                "'{}' should be an invalid protein seq".format(c),
            )

    def test_exact_search(self):
        # Test for 10 random
        for _ in range(10):
            i = random.randint(0, len(self.db.db.root.Protein.Entries))
            enr = i + 1
            s = self.db.get_sequence(enr)
            self.assertTrue(
                (enr in set(self.db.seq_search.search(s, is_sanitised=True)[1])),
                "exact search for entry {} failed.".format(i),
            )

    def get_random_subsequence(self, minlen=10):
        i = random.randint(0, len(self.db.db.root.Protein.Entries))
        elen = self.db.db.root.Protein.Entries[i]["SeqBufferLength"] - 1
        enr = i + 1

        ii = random.randint(0, elen - minlen)
        jj = random.randint(ii + minlen, elen)

        s = self.db.get_sequence(enr)[ii:jj]
        return s, enr, ii, jj

    def test_approx_search(self):
        # Test for random subsequence of 10 random sequences.
        for _ in range(10):
            s, enr, start_idx, end_idx = self.get_random_subsequence()
            approx_search_results = self.db.seq_search.approx_search(
                s, is_sanitised=True
            )
            self.assertIn(
                enr,
                {z[0] for z in approx_search_results},
                "approx search for entry {}[{}:{}] failed.".format(
                    enr - 1, start_idx, end_idx
                ),
            )

    def test_specific_approx_search_that_failed_on_jenkins(self):
        # TODO fix this problem!
        enrs = [15885, 16452]
        ranges = [(39, 376), (55, 140)]
        for enr, (start_idx, end_idx) in zip(enrs, ranges):
            seq = self.db.get_sequence(enr)[start_idx:end_idx]
            approx_search_results = self.db.seq_search.approx_search(
                seq, is_sanitised=True
            )
            enrs_with_approx_match = {z[0] for z in approx_search_results}
            self.assertIn(enr, enrs_with_approx_match)

    def test_map_to_hog(self):
        hog_mapper = ClosestSeqMapper(self.db)
        seqs = []
        entry_nrs = []
        for _ in range(10):
            s, enr, start_idx, end_idx = self.get_random_subsequence()
            entry_nrs.append(enr)
            seqs.append(
                SeqRecord(
                    Seq(s.decode()),
                    id=str(enr),
                    annotations={"molecule_type": "protein"},
                )
            )
        res = list(
            filter(
                lambda x: x.query == str(x.closest_entry_nr),
                hog_mapper.imap_sequences(seqs),
            )
        )
        self.assertEqual(10, len(res))
        for case in res:
            self.assertTrue(0 < case.distance < 1)

    def test_oma_group_from_numeric_id(self):
        group_id = 5
        grp = self.db.oma_group_members(group_id)
        self.assertEqual(4, len(grp))
        for e in grp:
            self.assertEqual(group_id, e["OmaGroup"])

    def test_fingerprint(self):
        fingerprint, grp_nr = "ESRTELL", 2617
        grp = self.db.oma_group_members(fingerprint)
        self.assertLessEqual(2, len(grp))
        for e in grp:
            self.assertEqual(grp_nr, e["OmaGroup"])

    def test_exon_structure(self):
        query = 14677  # Q8I237
        exons = self.db.get_exons(query)
        self.assertEqual(3, len(exons))

    def test_go_term_search(self):
        query = "GO:0004575"
        nrs = self.db.entrynrs_with_go_annotation(query, evidence="IDA")
        self.assertGreaterEqual(
            len(nrs), 1, "query GO term is known to occure at least in MAL32_YEAST"
        )
        for enr in nrs:
            self.assertIn(4575, self.db.get_gene_ontology_annotations(enr)["TermNr"])

    def test_induced_pairwise_orthologs(self):
        query = "YEAST3523"
        query_entry = self.db.ensure_entry(self.db.id_resolver.resolve(query))
        orthologs = self.db.get_hog_induced_pairwise_orthologs(query_entry)
        self.assertEqual(3, len(orthologs))
        self.assertCountEqual([b"1:1", b"1:1", b"m:1"], orthologs["RelType"])

    def test_induced_pairwise_paralogs(self):
        query = "YEAST12"
        query_entry = self.db.ensure_entry(self.db.id_resolver.resolve(query))
        orthologs = self.db.get_hog_induced_pairwise_paralogs(query_entry)
        self.assertEqual(1, len(orthologs))
        self.assertEqual(
            b"Saccharomyces cerevisiae (strain ATCC 204508 / S288c)",
            orthologs["DivergenceLevel"],
        )


class XRefDatabaseMock(Database):
    def __init__(self, name=None):
        if name is None:
            name = "xref.h5"
        f = tables.open_file(name, "w", driver="H5FD_CORE", driver_core_backing_store=0)
        entries = numpy.zeros(5, tables.dtype_from_descr(tablefmt.ProteinTable))
        entries["EntryNr"] = numpy.arange(1, 6)
        t = f.create_table("/Protein", "Entries", obj=entries, createparents=True)
        t.colinstances["EntryNr"].create_csindex()
        xref = numpy.zeros(10, tables.dtype_from_descr(tablefmt.XRefTable))
        xref["EntryNr"] = numpy.arange(1, 6, 0.5).astype(numpy.int32)
        xref["XRefSource"] = numpy.tile([0, 20], 5)
        xref["XRefId"] = ["XA{:03}g1.4".format(i) for i in range(10)]
        xref["Verification"] = tuple(itertools.islice(itertools.cycle([0, 2, 4]), 10))
        f.create_table("/", "XRef", tablefmt.XRefTable, obj=xref)
        f.create_group("/", "XRef_Index")
        for n in ("suffix", "buffer", "offset"):
            f.create_carray("/XRef_Index", n, obj=numpy.ones((5,), "i4"))
        self.db = f


class XRefIdMapperTest(unittest.TestCase):
    @classmethod
    def setUp(self):
        patch_db = XRefDatabaseMock()
        self.xrefmapper = XrefIdMapper(patch_db)

    def tearDown(self):
        self.xrefmapper._db.db.close()

    def test_multiple_xrefs_per_entry(self):
        xref_e1 = self.xrefmapper.map_entry_nr(1)
        self.assertEqual(len(xref_e1), 2)

    def test_map_many_entries(self):
        all_mapped = self.xrefmapper.map_many_entry_nrs(numpy.arange(1, 4))
        expected_len = (
            4 if isinstance(self.xrefmapper, XRefNoApproximateIdMapper) else 6
        )
        self.assertEqual((expected_len,), all_mapped.shape)
        self.assertEqual(self.xrefmapper.xref_tab.dtype, all_mapped.dtype)

    def test_map_entry_iterator(self):
        it = self.xrefmapper.iter_xrefs_for_entry_nr(1)
        self.assertTrue(isinstance(it, types.GeneratorType), "not a generator")
        exp_xrefs = ["XA000g1.4", "XA001g1.4"]
        for dic in it:
            self.assertIn(dic["xref"], exp_xrefs)

    def test_search_of_modified_xref(self):
        xref = "XA002g1.4"
        res = self.xrefmapper.search_xref(xref)
        self.assertEqual(2, res["EntryNr"])

    def test_verification_is_returned(self):
        res = self.xrefmapper.map_entry_nr(2)
        if isinstance(self.xrefmapper, XRefNoApproximateIdMapper):
            expected = ["exact"]
        else:
            expected = ["modified", "exact"]
        self.assertEqual(expected, [z["seq_match"] for z in res])


class XRefNoApproxIdMapperTest(XRefIdMapperTest):
    def setUp(self):
        patch_db = XRefDatabaseMock()
        self.xrefmapper = XRefNoApproximateIdMapper(patch_db)

    def test_modified_xref_not_returned_in_map(self):
        res = self.xrefmapper.map_entry_nr(2)
        xrefs = [x["xref"] for x in res]
        self.assertNotIn("XA002g1.4", xrefs)
        self.assertIn("XA003g1.4", xrefs)


class IdResolverTests(unittest.TestCase):
    @classmethod
    def setUp(self):
        patch_db = XRefDatabaseMock("test_idresolver.h5")
        self.xrefmapper = XrefIdMapper(patch_db)
        self.id_resolver = IDResolver(patch_db)
        patch_db.id_mapper = {"XRef": self.xrefmapper}

    def tearDown(self):
        self.xrefmapper._db.db.close()

    def test_search_of_modified_xref(self):
        xref = "XA002g1.4"
        res = self.id_resolver.search_xrefs(xref, return_seq_modified=True)
        self.assertEqual(2, res[0])
        self.assertTrue(res[1], "expected that {} is a modified sequence".format(xref))

    def test_search_of_unchecked_xref(self):
        xref = "XA001g1.4"
        res = self.id_resolver.search_xrefs(xref, return_seq_modified=True)
        self.assertEqual(1, res[0])
        self.assertFalse(
            res[1], "expected that {} is a unchecked sequence".format(xref)
        )


class TaxonomyTest(unittest.TestCase):
    tax_input = None
    maxDiff = None  # show complete strings if test fails

    @classmethod
    def setUpClass(cls):
        path = find_path_to_test_db("TestDb.h5")
        h5 = tables.open_file(path)
        cls.tax_input = h5.root.Taxonomy.read()
        h5.close()

    def setUp(self):
        self.tax = Taxonomy(self.tax_input)

    def test_parents(self):
        lin = [x["Name"].decode() for x in self.tax.get_parent_taxa(284811)]
        self.assertEqual(
            lin,
            [
                "Ashbya gossypii (strain ATCC 10895 / CBS 109.51 / FGSC 9923 / NRRL Y-1056)",
                "Eremothecium",
                "Saccharomycetaceae",
                "Saccharomycetales",
                "Saccharomycetes",
                "Saccharomycotina",
                "saccharomyceta",
                "Ascomycota",
                "Dikarya",
                "Fungi",
                "Opisthokonta",
                "Eukaryota",
            ],
        )

    def test_newick(self):
        member = frozenset(
            [
                self.tax._taxon_from_numeric(x)["Name"]
                for x in self.tax.tax_table["NCBITaxonId"]
            ]
        )
        phylo = self.tax.get_induced_taxonomy(member, collapse=True)
        expected = "(((Ashbya gossypii [strain ATCC 10895 / CBS 109.51 / FGSC 9923 / NRRL Y-1056],Saccharomyces cerevisiae [strain ATCC 204508 / S288c])Saccharomycetaceae,Schizosaccharomyces pombe [strain 972 / ATCC 24843])Ascomycota,Plasmodium falciparum [isolate 3D7])Eukaryota;"
        expected = expected.replace(" ", "_")
        self.assertEqual(expected, phylo.newick())

    def test_phylogeny(self):
        member = frozenset(
            [
                self.tax._taxon_from_numeric(x)["Name"]
                for x in self.tax.tax_table["NCBITaxonId"]
            ]
        )
        phylo = self.tax.get_induced_taxonomy(member, collapse=True)
        expected = {
            "id": 2759,
            "name": "Eukaryota",
            "children": [
                {
                    "id": 4890,
                    "name": "Ascomycota",
                    "children": [
                        {
                            "id": 4893,
                            "name": "Saccharomycetaceae",
                            "children": [
                                {
                                    "id": 284811,
                                    "name": "Ashbya gossypii (strain ATCC 10895 / CBS 109.51 / FGSC 9923 / NRRL Y-1056)",
                                },
                                {
                                    "id": 559292,
                                    "name": "Saccharomyces cerevisiae (strain ATCC 204508 / S288c)",
                                },
                            ],
                        },
                        {
                            "id": 284812,
                            "name": "Schizosaccharomyces pombe (strain 972 / ATCC 24843)",
                        },
                    ],
                },
                {"id": 36329, "name": "Plasmodium falciparum (isolate 3D7)"},
            ],
        }
        self.assertEqual(expected, phylo.as_dict())

    def test_induced_tax_simple_subtree(self):
        members = [559292, 284811]
        phylo = self.tax.get_induced_taxonomy(members)
        expected = {
            "id": 0,
            "name": "LUCA",
            "children": [
                {
                    "id": 284811,
                    "name": "Ashbya gossypii (strain ATCC 10895 / CBS 109.51 / FGSC 9923 / NRRL Y-1056)",
                },
                {
                    "id": 559292,
                    "name": "Saccharomyces cerevisiae (strain ATCC 204508 / S288c)",
                },
            ],
        }
        self.assertEqual(expected, phylo.as_dict())

    def test_induced_tax_with_parents_subtree(self):
        members = [559292, 284811]
        phylo = self.tax.get_induced_taxonomy(members, augment_parents=True)
        expected = {
            "id": 4893,
            "name": "Saccharomycetaceae",
            "children": [
                {
                    "id": 284811,
                    "name": "Ashbya gossypii (strain ATCC 10895 / CBS 109.51 / FGSC 9923 / NRRL Y-1056)",
                },
                {
                    "id": 559292,
                    "name": "Saccharomyces cerevisiae (strain ATCC 204508 / S288c)",
                },
            ],
        }
        self.assertEqual(expected, phylo.as_dict())

    def test_subtree_tax(self):
        clade = "Fungi"
        subtax = self.tax.get_subtaxonomy_rooted_at(clade)
        self.assertIn(4893, subtax.tax_table["NCBITaxonId"])
        self.assertNotIn(36329, subtax.tax_table["NCBITaxonId"])

    def test_taxid_of_extent_genomes(self):
        taxids = set(self.tax.get_taxid_of_extent_genomes())
        self.assertEqual({284811, 559292, 284812, 36329}, taxids)

    def test_induced_tax_with_gaps(self):
        pass


class TaxonomyTestInternalLevelSpecies(unittest.TestCase):
    """This test asserts that species A, which is also an internal level
    of the taxonomy is still reported as an external species via a
    <sciname> (disambiguate <code>) leaf."""

    taxtab = numpy.array(
        [(10, 0, b"Root"), (20, 10, b"Outgroup"), (30, 10, b"A"), (40, 30, b"B")],
        dtype=tables.dtype_from_descr(tablefmt.TaxonomyTable),
    )

    def setUp(self):
        patcher = mock.patch("pyoma.browser.models.Genome")
        self.addCleanup(patcher.stop)
        genome = patcher.start()
        type(genome).uniprot_species_code = mock.PropertyMock(return_value="HELLO")
        self.tax = Taxonomy(self.taxtab, genomes={20: genome, 30: genome, 40: genome})

    def test_newick(self):
        exp = "(Outgroup,(B,A_[disambiguate_HELLO])A)Root;"
        self.assertEqual(exp, self.tax.newick())

    def test_phyloxml(self):
        ins = (
            b"<scientific_name>A (disambiguate HELLO)</scientific_name>",
            b"<scientific_name>A</scientific_name>",
            b"<scientific_name>B</scientific_name>",
            b"<scientific_name>Outgroup</scientific_name>",
        )
        outs = (
            b"<scientific_name>B (disambiguate",
            b"<scientific_name>Outgroup (disambiguate",
        )
        res = self.tax.as_phyloxml()
        for part in ins:
            self.assertIn(part, res)
        for part in outs:
            self.assertNotIn(part, res)

    def test_dict_repr(self):
        res = self.tax.as_dict()
        self.assertDictEqual(
            {
                "name": "Root",
                "id": 10,
                "children": [
                    {"name": "Outgroup", "id": 20, "code": "HELLO"},
                    {
                        "name": "A",
                        "id": 30,
                        "children": [
                            {"name": "B", "id": 40, "code": "HELLO"},
                            {
                                "name": "A (disambiguate HELLO)",
                                "id": 30,
                                "code": "HELLO",
                            },
                        ],
                    },
                ],
            },
            res,
        )

    def test_induced_subtree_retains_internal_species(self):
        phylo = self.tax.get_induced_taxonomy([20, 40], augment_parents=True)
        self.assertIn(30, phylo.tax_table["NCBITaxonId"])


class DBMock(object):
    def __init__(self, h5):
        self.h5 = h5

    def get_hdf5_handle(self):
        return self.h5


class MockDBTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        path = find_path_to_test_db("TestDb.h5")
        cls.db = DBMock(tables.open_file(path))

    @classmethod
    def tearDownClass(cls):
        cls.db.get_hdf5_handle().close()


class GenomeIdResolverTest(MockDBTestCase):
    def setUp(self):
        self.OmaIdMapper = OmaIdMapper(self.db)

    def test_resolve_uniprot_code(self):
        query = "YEAST"
        g = self.OmaIdMapper.identify_genome(query)
        self.assertEqual(g["UniProtSpeciesCode"], query.encode("ascii"))

    def test_resolve_non_existing_uniprot_code(self):
        query = "XED22"
        with self.assertRaises(UnknownSpecies):
            self.OmaIdMapper.identify_genome(query)

    def test_resolve_taxon_id(self):
        expected = b"PLAF7"
        for query in (36329, b"36329", "36329"):
            g = self.OmaIdMapper.identify_genome(query)
            self.assertEqual(
                g["UniProtSpeciesCode"],
                expected,
                "failed for {} (type {})".format(query, type(query)),
            )

    def test_resolve_nonexisting_code(self):
        with self.assertRaises(UnknownSpecies):
            self.OmaIdMapper.identify_genome(2)

    def test_approx_search_genome(self):
        query = "sacero cervesa"
        expect = "YEAST"
        cands = self.OmaIdMapper.approx_search_genomes(query)
        self.assertIn(expect, [g.uniprot_species_code for g in cands])


class TestPerGenomeMetaData(MockDBTestCase):
    def setUp(self) -> None:
        self.pg = PerGenomeMetaData(self.db.get_hdf5_handle(), "YEAST")

    def test_in_oma_groups_matches(self):
        pass


class FastMapperTester(unittest.TestCase):
    def setUp(self) -> None:
        self.fast_mapper = FastMapper(Database(find_path_to_test_db()))

    def tearDown(self) -> None:
        self.fast_mapper.db.close()

    def test_search_small_sequence(self):
        test_seq = io.StringIO(">test\nAS")
        seq_iter = SeqIO.parse(test_seq, format="fasta")
        with self.assertLogs("pyoma", level=logging.INFO) as cm:
            anno = list(self.fast_mapper.iter_projected_goannotations(seq_iter))
        self.assertIn("Skipping", "\n".join(cm.output))
        self.assertEqual(0, len(anno))

    def test_search_existing_sequance(self):
        pass



from builtins import int, bytes, str
import collections
import csv
import logging
import math
import re
import numpy


"""
IMPORTANT NOTE:
---------------
This module has been copied from the dessimoz zoo library
directly. If you want to add functionality to this module,
make sure it is also integrated into the zoo. At the moment
we don't want to depend with pyoma on the zoo library as it
has many dependencies that are difficult to maintain.


Gene ontology module defining classes and methods to parse
and navigate the gene ontology DAG as well as to parse GO
annotations.


:author: Adrian Altenhoff
:institute: ETH Zurich
"""

NUM_ONT = 3
NUM_GO_ID_DIGITS = 7
UP_RELS = frozenset(["is_a", "part_of"])
_REV_RELS = {"is_a": "can_be", "part_of": "has_part"}


def reverse_name_of_rels(rels):
    down = frozenset([_REV_RELS[z] for z in rels])
    return down


def validate_go_id(term):
    if isinstance(term, (int, numpy.integer)):
        return int(term)

    term = term.strip()
    if term.startswith("GO:"):
        digits = term[3:]
    else:
        digits = term
    if not digits.isdigit() or len(digits) > NUM_GO_ID_DIGITS:
        raise ValueError("GO ID {} is not a valid go term".format(term))
    return int(digits)


class GOAspect(object):
    aspects = dict(molecular_function=0, biological_process=1, cellular_component=2)
    aspect2char = {0: "F", 1: "P", 2: "C"}

    @classmethod
    def from_string(cls, aspect):
        return cls.aspects[aspect]

    @classmethod
    def to_string(cls, aspectnr):
        for o, i in cls.aspects.items():
            if i == aspectnr:
                return o
        raise KeyError("aspect number not found: " + str(aspectnr))

    @classmethod
    def to_char(cls, aspectnr):
        # Converts an encoded aspect to the character required for GOA files
        return cls.aspect2char[aspectnr]


class GOterm(object):
    """A class representing a single Gene Ontology term.

    This class can serve as a factory for the OntologyParser. For that,
    pass it as a factory on 'Term'. """

    def __init__(self, stanza):
        self.id = validate_go_id(stanza["id"][0])
        self.name = " ".join(stanza["name"])
        self.definition = " ".join(stanza["def"])
        self.aspect = GOAspect.from_string(" ".join(stanza["namespace"]))
        self.is_a = [validate_go_id(parent) for parent in stanza["is_a"]]
        for rel in stanza["relationship"]:
            reltype, partner = rel.strip().split()
            if not reltype in self.__dict__.keys():
                self.__dict__[reltype] = list()
            self.__dict__[reltype].append(validate_go_id(partner))

    def replace_parentnames_by_refs(self, ont):
        for rel in [("is_a", "can_be"), ("part_of", "has_part")]:
            if rel[0] in self.__dict__.keys():
                for i, parent_id in enumerate(self.__dict__[rel[0]]):
                    parent_obj = ont[parent_id]
                    self.__dict__[rel[0]][i] = parent_obj
                    parent_obj._add_relation(self, rel[1])

    def _add_relation(self, term, rel):
        if rel not in self.__dict__.keys():
            self.__dict__[rel] = list()
        self.__dict__[rel].append(term)

    def get_parents(self, rels=None):
        """iterate over the direct parent GO terms.

        by default "is_a" and "part_of" relations are followed. This can be overwritten
        with the `rels`.

        :param rels: a set of relations to follow."""
        if rels is None:
            rels = UP_RELS
        for rel in rels:
            try:
                for term in getattr(self, rel):
                    yield term
            except AttributeError:
                pass

    def __str__(self):
        fmt = "GO:{{0:0{}d}}".format(NUM_GO_ID_DIGITS)
        return fmt.format(self.id)


class AbstractParser(object):
    def __init__(self, fp):
        self.close_fp = False
        if isinstance(fp, str):
            if fp.endswith(".gz"):
                from gzip import GzipFile

                fp = GzipFile(fp, "rb")
            else:
                fp = open(fp, "r")
            self.close_fp = True
        self.fp = fp
        self._read_headers()

    def _read_headers(self):
        pass

    def close_if_opened(self):
        if self.close_fp:
            self.fp.close()


class OntologyParser(AbstractParser):
    """A general purpose Ontolgoy parser

    Any ontology in the OBO format can be parsed with this object. The
    stanzas are converted to objects using the factories passed in the
    initializer."""

    def __init__(self, fp, factories=None):
        """creates an ontology parser

        :param fp: a filehandle or path to file (either plaintext or
            gzipped) containing the ontology.
        :param factories: a dictionary containing per stanza class
            (e.g. [Term]) a factory that returns an object from the
            data. The data is passed as dict to the factory"""
        if not factories:
            factories = dict(Term=GOterm)
        super(OntologyParser, self).__init__(fp)
        self.factories = factories
        self.tag_value_pair_re = re.compile(r"\s*(?P<tag>[^:]+):\s*(?P<value>[^!]*)")
        self.stanza_name_re = re.compile(r"\[(?P<name>[^]]*)\]")

    def stanzas(self):
        """iterates over the stanzas in the ontology yielding
        objects according to the factory parameter provided
        in the constructor."""
        curStanza = None
        for line in self.fp:
            line = line.strip()
            if not line or line[0] == "!":
                continue

            # check whether a new stanza starts
            match = self.stanza_name_re.match(line)
            if match is not None:
                obj = self._create_obj_from_stanza(curStanza)
                if obj is not None:
                    yield obj
                curStanza = collections.defaultdict(list)
                curStanza["_name"] = match.group("name")
            elif curStanza is not None:
                # it has to be value-pair line. add it to the stanza
                match = self.tag_value_pair_re.match(line)
                curStanza[match.group("tag")].append(match.group("value"))
        obj = self._create_obj_from_stanza(curStanza)
        if obj is not None:
            yield obj
        self.close_if_opened()

    def _create_obj_from_stanza(self, stanza):
        """method which creates the appropriate object from a given
        stanza or returns None, e.g. for stanza types without a provided
        factory or obsolete terms or ..."""
        res = None
        if stanza is not None:
            try:
                factory = self.factories[stanza["_name"]]
            except KeyError:
                # no factory for this stanza type. ignore
                pass
            else:
                if (not "is_obsolete" in stanza) or (
                    not "true" in stanza["is_obsolete"]
                ):
                    res = factory(stanza)
        return res

    def __iter__(self):
        return self.stanzas()


class GeneOntology(object):
    """The GeneOntology object contains the whole ontology in an internal
    format and allows to traverse it efficiently."""

    def __init__(self, parser, rels=None):
        if rels is None:
            rels = UP_RELS
        if not isinstance(parser, OntologyParser):
            raise Exception("requires an OntologyParser instance")
        self.parser = parser
        self.up_rels = UP_RELS.intersection(rels)
        self.down_rels = reverse_name_of_rels(self.up_rels)

    def parse(self):
        """parse the ontology data.

        This method should be called after instantiation and before any traversal"""
        self.terms = dict()
        for cur_term in self.parser:
            self.terms[cur_term.id] = cur_term

        # replace parents nrs by the references to the GO terms objects
        # this can be only done once all the terms have been created
        for term in self.terms.values():
            term.replace_parentnames_by_refs(self.terms)

    def ensure_term(self, term):
        """returns the term object associated with term. if term is already
        a GOterm object, it is simply return. term_ids can be either numeric
        ids of existing GO-terms or propper GO-terms ids, i.e. GO:000002

        :param term: a term_id or a GOterm object"""
        if isinstance(term, GOterm):
            return term
        else:
            return self.term_by_id(term)

    def term_by_id(self, term_id):
        """Returns the term object associated with term_id.

        :param term_id: a GO-term number or a GO-term id (GO:008150)."""
        try:
            term = self.terms[validate_go_id(term_id)]
            return term
        except KeyError:
            raise ValueError(str(term_id) + " is an invalid GO term.")

    def get_superterms_incl_queryterm(self, term, max_steps=-1):
        """returns a set with all the superterms of a query term.

        :param max_steps: The search can be limited to contain only
            terms that are at most 'max_steps' upwards. If set to -1, no
            limit is applied and the search goes up to the root."""
        term = self.ensure_term(term)
        return self._traverseGraph(term, max_steps, self.up_rels)

    def get_subterms(self, term, max_steps=-1):
        term = self.ensure_term(term)
        return self._traverseGraph(term, max_steps, self.down_rels)

    def _traverseGraph(self, node, max_steps, rels):
        """_traverseGraph traverses the graph in a breath first manner
        and reports all the nodes reachable within max_steps."""
        remain = set([node])
        found = set()
        while len(remain) > 0 and max_steps != 0:
            novel = set()
            for t in remain:
                for rel in rels:
                    try:
                        novel.update(t.__dict__[rel])
                    except KeyError:
                        pass
            found.update(remain)
            remain = novel.difference(found)
            max_steps -= 1
        return found


class FreqAwareGeneOntology(GeneOntology):
    """GO hierarchy represents the Gene Ontology vocabulary.

    It gets loaded from the xml file and, in conjunction with
    an annotation file (GOA) the relative frequencies per term get
    estimated. this estimation respects the hierarchy of the
    vocabulary.
    Further, this class provides methods to traverse the hierarchy
    in an easy way."""

    def __init__(self, fp, rels=UP_RELS):
        super(FreqAwareGeneOntology, self).__init__(fp, rels=rels)
        self.reset_freqs()

    def reset_freqs(self):
        self.cnts = dict()
        self.tot_cnts = [0] * NUM_ONT

    def estimate_freqs(self, annotations):
        for anno in annotations:
            try:
                self._update_counts(self.term_by_id(anno["TermNr"]))
            except ValueError:
                logging.info(
                    "invalid annotation term_id in freq estim:" + str(anno.term_id)
                )

    def _update_counts(self, term):
        for cur_term in self.get_superterms_incl_queryterm(term):
            self.cnts[cur_term.id] = self.cnts.get(cur_term.id, 0) + 1
        self.tot_cnts[cur_term.aspect] += 1

    def get_term_frequency(self, term):
        term = self.ensure_term(term)
        try:
            freq = self.cnts.get(term.id, 0) / self.tot_cnts[term.aspect]
            return freq
        except ZeroDivisionError:
            return 0

    def last_common_ancestor(self, *terms):
        cand = self.get_superterms_incl_queryterm(terms[0])
        for t in terms[1:]:
            cand.intersection_update(self.get_superterms_incl_queryterm(t))
        lca = min(cand, key=self.get_term_frequency)
        return lca

    def lin_similarity(self, term1, term2):
        term1 = self.ensure_term(term1)
        term2 = self.ensure_term(term2)
        if term1.aspect != term2.aspect:
            # early abort, since the two terms will be by
            # definition not similar
            sim = 0
        else:
            lca = self.last_common_ancestor(term1, term2)
            sim = (
                2
                * math.log(self.get_term_frequency(lca))
                / (
                    math.log(self.get_term_frequency(term1))
                    + math.log(self.get_term_frequency(term2))
                )
            )
        return sim


class AnnotationFilter(object):
    EXP_CODES = frozenset(["EXP", "IDA", "IPI", "IMP", "IGI", "IEP"])
    TRUST_IEA_REFS = frozenset(
        [
            "GO_REF:0000002",
            "GOA:interpro",
            "GOA:interpro|GO_REF:0000002",  # InterPro
            "GO_REF:0000003",
            "GOA:spec",
            "GOA:spec|GO_REF:0000003" "GO_REF:0000004",  # EC number
            "GOA:spkw",
            "GOA:spkw|GO_REF:0000004",
            "GO_REF:0000037",
            "GO_REF:0000038" "GO_REF:0000023",  # SwissProt Keywords
            "GOA:spsl",
            "GOA:spsl|GO_REF:0000023",
            "GO_REF:0000039",
            "GO_REF:0000040",  # UniProtKB Subcellular Location
        ]
    )

    @staticmethod
    def is_negated(a):
        return a.qualifier.find("NOT") >= 0

    @classmethod
    def is_exp_annotation(cls, a):
        return a.evidence in cls.EXP_CODES and not cls.is_negated(a)

    @classmethod
    def is_trusted_electronic(cls, a):
        return (
            a.evidence == "IEA"
            and a.db_ref in cls.TRUST_IEA_REFS
            and not cls.is_negated(a)
        )

    @classmethod
    def is_exp_or_trusted_electronic(cls, a):
        return cls.is_exp_annotation(a) or cls.is_trusted_electronic(a)


# definition of the GOA Annotations. for performance reasons, we
# keep this as a namedtuple collection.
GOA_Annotation = collections.namedtuple(
    "GOA_Annotation",
    [
        "db",
        "db_obj_id",
        "db_obj_sym",
        "qualifier",
        "term_id",
        "db_ref",
        "evidence",
        "with_from",
        "aspect",
        "db_obj_name",
        "db_obj_syn",
        "db_obj_typ",
        "taxon",
        "date",
        "assigned_by",
        "ext",
        "gene_product_from_id",
    ],
)


class AnnotationParser(object):
    def __init__(self, fp, factory=GOA_Annotation._make):
        self._needs_close = False
        if isinstance(fp, str):
            if fp.endswith(".gz"):
                from gzip import GzipFile

                fp = GzipFile(fp, "rb")
                self._needs_close = True
            else:
                fp = open(fp, "rb")
        self.fp = fp
        self.factory = factory

        self._read_headers()

    def _read_headers(self):
        pass

    def annotations(self):
        """Iterates over the annotations in the file yielding objects
        constructed by the factory argument passed to the constructor
        of this class for each annotation."""
        csv_reader = csv.reader(
            (l for l in self.fp if not l.startswith("!")), delimiter="\t"
        )
        for row in csv_reader:
            yield self.factory(row)
        if self._needs_close:
            self.fp.close()

    def __iter__(self):
        return self.annotations()



from functools import wraps
import logging
from time import time


def timethis(level, name=None):
    """
    Decorator that reports the execution time. level is the logging
    level, name is the logger name. If name isn't specified,
    it default to the function's module

    :Example:
    @timethis(logging.INFO)
    def add(x,y):
        return x+y

    This would log the execution time with level info to the
    module's logger
    """

    def decorate(func):
        logname = name if name else func.__module__
        log = logging.getLogger(logname)

        @wraps(func)
        def wrapper(*args, **kwargs):
            start = time()
            result = func(*args, **kwargs)
            end = time()
            log.log(level, "time for {}: {}sec".format(func.__name__, end - start))
            return result

        return wrapper

    return decorate



import itertools

import networkx as nx
import logging
import numpy
import pyham
import tables
from .tablefmt import AncestralSyntenyRels
from . import db
from .models import Genome
from concurrent.futures import TimeoutError
import multiprocessing as mp
from pebble import ProcessExpired, concurrent

logger = logging.getLogger(__name__)
mp.set_start_method("fork")


def assign_extant_syteny(h5, ham):
    genomes = [Genome(h5, g) for g in h5.get_hdf5_handle().root.Genome.read()]
    for g in genomes:
        proteins = h5.main_isoforms(g.uniprot_species_code)
        proteins.sort(order=["Chromosome", "LocusStart"])
        graph = nx.Graph()
        gene1 = ham.get_gene_by_id(proteins[0]["EntryNr"])
        graph.add_node(gene1)
        for i in range(1, len(proteins)):
            gene = ham.get_gene_by_id(proteins[i]["EntryNr"])
            graph.add_node(gene)
            if proteins[i - 1]["Chromosome"] == proteins[i]["Chromosome"]:
                # connect genes only if they are on the same chromosome
                graph.add_edge(gene1, gene, weight=1)
            gene1 = gene
        if logger.isEnabledFor(logging.INFO):
            logger.info(
                "Synteny for extant genome {} established. |V|={}, |E|={}, |CC|={}".format(
                    g.uniprot_species_code,
                    graph.number_of_nodes(),
                    graph.number_of_edges(),
                    nx.number_connected_components(graph),
                )
            )
        gene1.genome.taxon.add_feature("synteny", graph)


def assign_ancestral_synteny(ham):
    for tree_node in ham.taxonomy.tree.traverse("postorder"):
        try:
            genome = tree_node.genome
        except AttributeError:
            logger.warning("No genome stored for {}".format(tree_node.name))
            continue

        if isinstance(tree_node.genome, pyham.ExtantGenome):
            logger.info(
                "synteny for extant genome '{}' already assigned".format(
                    tree_node.genome
                )
            )
            continue
        elif isinstance(tree_node.genome, pyham.AncestralGenome):
            graph = nx.Graph()
            graph.add_nodes_from(tree_node.genome.genes)
            for child in tree_node.children:
                # we remove nodes that do not have a parent hog and connect their neighbors
                # (given they have exactly two neighbors) in a working copy that is used
                # to propagate the synteny graph to the higher level.
                G = child.synteny.copy()
                remove_nodes_on_path_without_parents(G)
                # cleanup graph if they get too big
                delete_irrelevant_edges(G, min_importance=7, max_neighbors=8)
                for u, v, weight in G.edges.data("weight", default=1):
                    if (
                        u.parent is not None
                        and v.parent is not None
                        and u.parent != v.parent
                    ):
                        assert u.parent in graph.nodes
                        assert v.parent in graph.nodes
                        parent_edge = (u.parent, v.parent)
                        if graph.has_edge(*parent_edge):
                            graph[parent_edge[0]][parent_edge[1]]["weight"] += weight
                        else:
                            graph.add_edge(*parent_edge, weight=weight)
            logger.info("build graph for {}: {}".format(tree_node.name, nx.info(graph)))
            remove_forks_from_gene_losses(graph)
            tree_node.add_feature("synteny", graph)
            logger.info(
                "Synteny for ancestral genome '{}' created. |V|={}, |E|={}, |CC|={}".format(
                    tree_node.name,
                    graph.number_of_nodes(),
                    graph.number_of_edges(),
                    nx.number_connected_components(graph),
                )
            )
            yield tree_node.name, graph


def remove_nodes_on_path_without_parents(G: nx.Graph):
    """Modifies inplace the Graph such that all nodes
    that have exactly two neighbors and do not have a parent
    HOG are removed and their adjacent nodes are reconnected.

    This works transitively, eg. if A - b - c - D is a graph,
    and b, c have no parents, the pruned graph will be A - D.

    :Example:
    >>> class N:
    ...    def __init__(self, parent=None):
    ...        self.parent = parent
    ...    def __repr__(self):
    ...        return "N({})".format(self.parent)
    >>> A = N('A')
    >>> b = N()
    >>> c = N()
    >>> D = N('D')
    >>> E = N('E')
    >>> G = nx.Graph()
    >>> G.add_weighted_edges_from([(A, b, 1), (b, c, 1), (c, D, 2), (D, E, 1)])
    >>> remove_nodes_on_path_without_parents(G)
    >>> G.edges
    EdgeView([(N(A), N(D)), (N(D), N(E))])

    :param nx.Graph G: input Graph to be analysed
    """
    cnt_removed_nodes = 0
    for node in list(G.nodes):  # convert to list to allow graph modification
        if node.parent is None and len(G.adj[node]) == 2:
            left, right = tuple(G.adj[node])
            if not G.has_edge(left, right):
                G.add_edge(
                    left,
                    right,
                    weight=max(G[left][node]["weight"], G[right][node]["weight"]),
                )
                G.remove_node(node)
                cnt_removed_nodes += 1
    logger.info(
        "removed {} nodes that have synteny but no parent hog.".format(
            cnt_removed_nodes
        )
    )


def delete_irrelevant_edges(G: nx.Graph, min_importance=10, max_neighbors=10):
    """removes edges that are no longer relevant.

    :Example:
    >>> G = nx.Graph()
    >>> G.add_weighted_edges_from([(0, 1, 4), (1, 9, 5)])
    >>> G.add_weighted_edges_from([(1, x, 1) for x in range(2,8)])
    >>> S = G.copy()
    >>> delete_irrelevant_edges(G, min_importance=3)
    >>> G.edges.data('weight')
    EdgeDataView([(0, 1, 4), (1, 9, 5)])
    >>> delete_irrelevant_edges(S, max_neighbors=5, min_importance=5)
    >>> S.edges.data('remove', 0)
    EdgeDataView([(0, 1, 0), (1, 9, 0), (1, 2, 1), (1, 3, 1), (1, 4, 1), (1, 5, 1), (1, 6, 1), (1, 7, 1)])

    """
    remove_edge_under_weight = max(min_importance - 2, 1.001)
    cnt_removed_edges = 0
    for u, nbrs in G.adjacency():
        weights = numpy.array([nbr["weight"] for nbr in nbrs.values()])
        weights.sort()
        pos = weights.searchsorted(min_importance)
        to_remove = []
        if pos <= len(weights) - 2:
            # we have at least two well supported edges. remove the ones
            # that are weakly supported.
            for v, eattr in nbrs.items():
                if eattr["weight"] < remove_edge_under_weight:
                    to_remove.append((u, v))
        elif len(weights) > max_neighbors:
            # mark all edge as potential removal, if marked from both sides, remove edge
            for v, eattr in nbrs.items():
                if eattr["weight"] < remove_edge_under_weight:
                    if "remove" in eattr:
                        to_remove.append((u, v))
                    else:
                        eattr["remove"] = 1
        G.remove_edges_from(to_remove)
        cnt_removed_edges += len(to_remove)
    logger.info("removed {} irrelevant edges".format(cnt_removed_edges))


def remove_forks_from_gene_losses(G: nx.Graph):
    """Remove forks in the graph that originate from gene losses
    on the child level.

    <h_a>  --  <h_i>  --  <h_j>  -- <h_b>
                 \           /
                  -- <h_k> --

    this is the base case to remove the edge between h_i and h_j.
    <h_k> has to be at least one hog, but can also be more, but
    the path from h_i to h_j via h_k needs to be simple, e.g. no
    further branching.
    """
    logger.info("starting remove_fork_from_gene_losses...")
    deg_3_nodes = (n[0] for n in G.degree if n[1] == 3)
    pot_cases = list(
        filter(lambda uv: G.has_edge(*uv), itertools.combinations(deg_3_nodes, 2))
    )
    removed_forks = 0
    for u, v in pot_cases:
        pgen = nx.shortest_simple_paths(G, u, v)
        direct = next(pgen)  # direct edge between u,v. to be removed
        try:
            next_path = next(pgen)
        except StopIteration:
            # no other path, continue with next pot_case
            continue
        try:
            more_path = next(pgen)
            continue  # we don't want cases that have more possible paths
        except StopIteration:
            pass

        # remove direct path, reweight next_path
        w = G.edges[u, v]["weight"]
        G.remove_edge(u, v)
        for i in range(len(next_path) - 1):
            G.edges[next_path[i], next_path[i + 1]]["weight"] += w
        removed_forks += 1
        logger.debug(
            "removed edge from ({},{}), found other path of len {}: {}".format(
                u, v, len(next_path), next_path
            )
        )
    logger.info(
        "removed {} forks in G(|V|={},|E|={})".format(
            removed_forks, G.number_of_nodes(), G.number_of_edges()
        )
    )
    logger.info("finished remove_fork_from_gene_losses")


def extract_hog_row_links(graph, hogid_2_hogrow_lookup, level):
    def map_id(id):
        try:
            k = id.index("_")
            query = id[:k]
        except ValueError:
            query = id
        return hogid_2_hogrow_lookup[query]

    edges = []
    nr_errors = 0
    for u, v, w in graph.edges.data("weight", default=1):
        try:
            edges.append((map_id(u), map_id(v), w))
        except KeyError as e:
            nr_errors += 1
            logger.error(
                "unmappable relation ({}, {}) on level {}: {}".format(u, v, level, e)
            )
    if nr_errors > 0:
        logger.warning("{} errors on level {}".format(nr_errors, level))
    return numpy.array(edges, dtype=tables.dtype_from_descr(AncestralSyntenyRels()))


def taxid_from_level(h5, level):
    try:
        taxnode = h5.tax.get_taxnode_from_name_or_taxid(level)
        taxid = int(taxnode["NCBITaxonId"])
    except KeyError as e:
        if level == "LUCA":
            taxid = 0
        else:
            raise
    logger.info("level '{}' -> taxid = {}".format(level, taxid))
    return taxid


@concurrent.process(timeout=1200)
def hogid_2_rownr(h5, level):
    lookup = {}
    level_query = "Level == {!r}".format(level.encode("utf-8"))
    row_iter = h5.get_hdf5_handle().root.HogLevel.where(level_query)
    for row in row_iter:
        lookup[row["ID"].decode()] = row.nrow
    logger.info("hogmap: found {} hog at level {}".format(len(lookup), level))
    return lookup


def level_only_graph_with_hogid_nodes(graph: nx.Graph) -> nx.Graph:
    G = nx.relabel_nodes(graph, lambda n: n.hog_id)
    return G


def infer_synteny(orthoxml, h5name, tree):
    h5 = db.Database(h5name)
    ham = pyham.Ham(
        tree_file=tree,
        hog_file=orthoxml,
        tree_format="newick",
        use_internal_name=True,
        species_resolve_mode="OMA",
    )
    assign_extant_syteny(h5, ham)
    synteny_graphs_mapped = {}
    for cnt, (level, graph) in enumerate(assign_ancestral_synteny(ham), start=1):
        shallow_graph = level_only_graph_with_hogid_nodes(graph)
        future = hogid_2_rownr(h5, level)
        try:
            hogid_lookup = future.result()
            taxid = taxid_from_level(h5, level)
            edges = extract_hog_row_links(shallow_graph, hogid_lookup, level)
            synteny_graphs_mapped[taxid] = edges
        except TimeoutError as error:
            logger.error(
                "hogid_2_rownr took longer than %d seconds".format(error.args[1])
            )
        except ProcessExpired as error:
            logger.error("{}. Exit code: {}}".format(error, error.exitcode))
        except Exception as error:
            logger.exception("hogid_2_rownr raised {}".format(error))

        if cnt % 200 == 0:
            logger.info("closing and reopening {} file".format(h5name))
            h5.close()
            h5 = db.Database(h5name)
    h5.close()
    return ham, synteny_graphs_mapped


def plot_graph_neighborhood(G, node, steps=2):
    import matplotlib.pyplot as plt

    nodes = [node] + [v for u, v in nx.bfs_edges(G, source=node, depth_limit=steps)]
    S = G.subgraph(nodes)
    colors = [w for u, v, w in S.edges.data("weight", 1)]
    weights = [w for u, v, w in S.edges.data("weight", 1)]
    options = {
        "node_color": "#A0CBE2",
        "edge_color": colors,
        "width": weights,
        "edge_cmap": plt.cm.jet,
        "with_labels": True,
    }
    pos = nx.spring_layout(S, iterations=200)
    nx.draw(S, pos=pos, **options)
    labels = nx.get_edge_attributes(S, "weight")
    nx.draw_networkx_edge_labels(S, pos=pos, edge_labels=labels)
    plt.show()


def write_syntenygraphs_to_hdf5(h5name, hog_edges_per_level):
    with tables.open_file(h5name, "a") as h5:
        node = h5.create_group(
            "/",
            "AncestralSynteny",
            title="Graph of HOGs (identified by rowidx in HogLevel table). Per tax level",
        )
        for level, edges in hog_edges_per_level.items():
            h5.create_table(
                node, "tax{}".format(level), obj=edges, expectedrows=len(edges)
            )


if __name__ == "__main__":
    orthoxml = "/Volumes/TOSHIBA EXT/basf/All.May2020/downloads/oma-hogs.orthoXML.gz"
    tree = "/Volumes/TOSHIBA EXT/basf/All.May2020/downloads/speciestree.nwk"
    h5db = "/Volumes/TOSHIBA EXT/basf/All.May2020/data/OmaServer.h5"
    infer_synteny(orthoxml, h5db, tree)



import tables


class GenomeFmt(tables.IsDescription):
    NCBITaxonId = tables.UInt32Col(pos=0)
    UniProtSpeciesCode = tables.StringCol(5, pos=1)
    TotEntries = tables.UInt32Col(pos=2)
    TotAA = tables.UInt32Col(pos=3)
    EntryOff = tables.UInt32Col(pos=4)
    SciName = tables.StringCol(255, pos=5)
    Release = tables.StringCol(255, pos=6)


class TaxonomyFmt(tables.IsDescription):
    NCBITaxonId = tables.UInt32Col(pos=0)
    ParentTaxonId = tables.UInt32Col(pos=1)
    Name = tables.StringCol(255, pos=2)


class BestMatchesFmt(tables.IsDescription):
    EntryNr1 = tables.UInt32Col(pos=0)
    EntryNr2 = tables.UInt32Col(pos=1)
    Score = tables.Float32Col(pos=2)
    PamDistance = tables.Float32Col(pos=3)
    PamVariance = tables.Float32Col(pos=4)
    SumLengths = tables.UInt16Col(pos=5)
    isSP = tables.BoolCol(pos=6, dflt=True)
    isVP = tables.BoolCol(pos=7, dflt=False)
    isCanonicalSplicing = tables.BoolCol(pos=8, dflt=True)



import ete3
import copy
import collections
import itertools


def get_newick_tree_from_tax_db(tax):
    def traverse(node):
        yield node
        if "children" in node:
            for child in node["children"]:
                yield from traverse(child)

    def rec(node):
        if "children" not in node:
            return str(node["id"])
        children = []
        for child in node["children"]:
            children.append(rec(child))
        return "({}){}".format(",".join(children), node["id"])

    def get_duplicates(node):
        c = collections.Counter(x["id"] for x in traverse(node))
        return list(
            item[0] for item in itertools.takewhile(lambda x: x[1] > 1, c.most_common())
        )

    def rename_internal_duplicates(node, duplicates):
        for n in traverse(node):
            if n["id"] in duplicates and "children" in n:
                n["id"] = "{}Rep".format(n["id"])

    taxdict = tax.as_dict()
    dupl = get_duplicates(taxdict)
    rename_internal_duplicates(taxdict, dupl)
    res = rec(taxdict) + ";"
    return res


def filter_tree(tree: ete3.PhyloNode, root_node=None, taxfilter=None):
    newtree = copy.deepcopy(tree)
    if root_node:
        for n in newtree.traverse():
            if str(n.name) == str(root_node):
                newtree = n
                break
    if taxfilter:
        for n in newtree.traverse():
            if n.name in taxfilter:
                n.detach()
    return newtree


def taxa_index(tree: ete3.PhyloNode):
    """build a dictionary from tree-node name to idx"""
    index = {n.name: i for i, n in enumerate(tree.traverse())}
    return index


def leaf_index(tree: ete3.PhyloNode):
    index = {n.name: i for i, n in enumerate(tree.iter_leaves())}
    return index



import logging
import numpy

logger = logging.getLogger(__name__)


class HogLevelFilter(object):
    """Filter sub-hogs at given level (or below)

    This class allows to identify  at a given level or - if a hog
    does not go deep enough to cover the desired level, the whole hog will be returned.
    """

    def __init__(self, db, level):
        self.db = db
        self.level = level
        tax_below_level = self.db.tax.get_subtaxonomy_rooted_at(level)
        self.sub_level_distances = self._build_distance_to_query_level_lookup(
            tax_below_level
        )
        self.parent_levels = set(
            db.tax.get_parent_taxa(
                db.tax.get_taxnode_from_name_or_taxid(level)[0]["NCBITaxonId"]
            )["Name"]
        )
        # cache all the families nrs that are defined in the clade of interest
        self.fams_in_clade = set(
            int(roothog["Fam"])
            for roothog in db.get_hdf5_handle().root.HogLevel.where(
                '~contains(ID, b".") & (IsRoot == True)'
            )
            if roothog["Level"] in self.sub_level_distances
            or roothog["Level"] in self.parent_levels
        )

    def _build_distance_to_query_level_lookup(self, tax):
        nodes = {}

        def traverse(node, dist=0):
            nodes[node["name"].encode("utf-8")] = dist
            try:
                for c in node["children"]:
                    traverse(c, dist + 1)
            except KeyError:
                pass

        traverse(tax.as_dict())
        return nodes

    def find_closest_children_level(self, levs):
        logger.debug(levs)
        min_dist, closest = 1000, None

        for lev in levs:
            try:
                if self.sub_level_distances[lev] < min_dist:
                    closest = lev.item().decode()
                    min_dist = self.sub_level_distances[lev]
            except KeyError:
                pass
        logger.debug("closest level {} at {}".format(closest, min_dist))
        return closest

    def analyse_families(self, it):
        for fam in it:
            if fam not in self.fams_in_clade:
                continue
            subhog_ids = self.db.get_subhogids_at_level(fam_nr=fam, level=self.level)
            if len(subhog_ids) == 0:
                # let's check if we have any level in the family that is a sub clade
                levs = self.db.hog_levels_of_fam(fam_nr=fam)
                closest_level = self.find_closest_children_level(levs)
                if closest_level is not None:
                    yield (self.db.format_hogid(fam), closest_level)
            else:
                for subhog_id in subhog_ids:
                    yield (subhog_id, self.level)


def compare_levels(
    parent_level_hogs: numpy.array,
    children_level_hogs: numpy.array,
    return_duplication_events=False,
):
    """compares hogs at two levels and returns duplicated/lost/gained/identical states

    The function requires that the hogs are sorted numpy arrays according to their HOGid"""
    annotated = []
    dups = 0
    i = j = 0
    while i < len(parent_level_hogs) and j < len(children_level_hogs):
        if parent_level_hogs[i]["Fam"] < children_level_hogs[j]["Fam"]:
            annotated.append(tuple(parent_level_hogs[i]) + (b"lost",))
            i += 1
        elif parent_level_hogs[i]["Fam"] > children_level_hogs[j]["Fam"]:
            annotated.append(tuple(children_level_hogs[j]) + (b"gained",))
            j += 1
        else:
            if parent_level_hogs[i]["ID"] == children_level_hogs[j]["ID"]:
                annotated.append(tuple(children_level_hogs[j]) + (b"retained",))
                i += 1
                j += 1
            else:
                dupl_fnd = 0
                while j < len(children_level_hogs) and children_level_hogs[j][
                    "ID"
                ].startswith(parent_level_hogs[i]["ID"]):
                    annotated.append(tuple(children_level_hogs[j]) + (b"duplicated",))
                    j += 1
                    dupl_fnd += 1
                if dupl_fnd > 0:
                    # parent duplicated into at least one subhog
                    dups += 1
                else:
                    # parent subhog does not exist, it is lost
                    annotated.append(tuple(parent_level_hogs[i]) + (b"lost",))
                i += 1

    while i < len(parent_level_hogs):
        annotated.append(tuple(parent_level_hogs[i]) + (b"lost",))
        i += 1
    while j < len(children_level_hogs):
        annotated.append(tuple(children_level_hogs[j]) + (b"gained",))
        j += 1
    dt = children_level_hogs.dtype.descr + [("Event", "S10")]
    diff = numpy.array(annotated, dtype=dt)
    if return_duplication_events:
        return diff, dups
    return diff



from xml.etree.ElementTree import XMLParser
from datasketch import MinHash, MinHashLSH
import gzip
import logging

logger = logging.getLogger(__name__)


class Comparer(object):
    def __init__(self):
        pass


class GeneMapper(object):
    attr = "protId"

    def map(self, attributes):
        return attributes[self.attr]


class FilteredSpeciesGeneMapper(GeneMapper):
    def __init__(self, skip_prefixes):
        prefix_lens = set(len(z) for z in skip_prefixes)
        if len(prefix_lens) != 1:
            raise ValueError("prefixes should be all equal size")
        self.skip_prefixes = skip_prefixes
        self.prefix_len = len(skip_prefixes[0])

    def map(self, attributes):
        if attributes[self.attr][: self.prefix_len] in self.skip_prefixes:
            return None
        else:
            return attributes[self.attrib]


class ToplevelOrthoXMLParser(object):
    def __init__(self, gene_mapper):
        self.gene_mapper = gene_mapper
        self.cur_hog_id = None
        self.cur_hog_depth = 0
        self.cur_hog_memb = []
        self.genes = {}
        self.levels = {}
        self.hogs = {}

    def start(self, tag, attrib):
        if tag == "{http://orthoXML.org/2011/}gene":
            self.genes[int(attrib["id"])] = self.gene_mapper.map(attrib)
        elif tag == "{http://orthoXML.org/2011/}geneRef":
            gene_id = int(attrib["id"])
            gene = self.genes[gene_id]
            if gene is not None:
                self.cur_hog_memb.append(gene)
        elif tag == "{http://orthoXML.org/2011/}orthologGroup":
            if self.cur_hog_depth == 0:
                self.cur_hog_id = int(attrib["id"])
            self.cur_hog_depth += 1
        elif (
            tag == "{http://orthoXML.org/2011/}property"
            and attrib["name"] == "TaxRange"
        ):
            if self.cur_hog_depth == 1:
                self.levels[self.cur_hog_id] = attrib["value"]

    def end(self, tag):
        if tag == "{http://orthoXML.org/2011/}orthologGroup":
            self.cur_hog_depth -= 1
            if self.cur_hog_depth == 0:
                self.hogs[self.cur_hog_id] = frozenset(self.cur_hog_memb)
                self.cur_hog_memb = []

    def data(self, data):
        pass

    def close(self):
        return


class Analysis(object):
    def __init__(self, hogs, rootlevels):
        self.hogs = hogs
        self.rootlevels = rootlevels


def load_toplevel_hogs(fpath, species_to_skip=None):
    open_ = gzip.open if fpath.endswith(".gz") else open
    with open_(fpath) as fh:
        if species_to_skip is None:
            gene_mapper = GeneMapper()
        else:
            gene_mapper = FilteredSpeciesGeneMapper(species_to_skip)
        toplevelparser = ToplevelOrthoXMLParser(gene_mapper)
        parser = XMLParser(target=toplevelparser)
        for line in fh:
            parser.feed(line)
    return Analysis(toplevelparser.hogs, toplevelparser.levels)



import tables
import numpy as np

try:
    from functools import lru_cache
except ImportError:
    from functools32 import lru_cache
from builtins import filter


class RelationsOfEntry(object):
    """
    This container class contains the relations from one given query
    gene in one other genome. E.g. all the BestMatches from HUMAN2 in
    MOUSE, or the StablePairs from CHICK4 in DROME... The different
    types of relations are in fact implemented as different
    subclasses.
    """

    @staticmethod
    def filter(x):
        if isinstance(x, np.ndarray):
            return np.array([True] * len(x), dtype="b")
        else:
            return np.array([True], dtype="b")

    def __init__(self, data):
        """create a RelationsOfEntry object from 'data'.
        data is assumed to be a numpy structured array containing
        matches from one query entry only, i.e. all EntryNr1 elements
        have the same value. Further, it is assumed that the data
        is ordered according to EntryNr2 (inc order)."""
        self.data = data

    def __iter__(self):
        return filter(self.filter, self.data)

    def __contains__(self, item):
        idx = np.where(self.data["EntryNr2"] == item)
        return self.filter(self.data[idx]).any()

    def relations(self):
        """return a copy of the matches restricted to the rows
        passing the filter criterion."""
        return self.data[self.filter(self.data)]

    def set_relations(self, value):
        if isinstance(value, (list, set, int)):
            value = np.array(value)
        elif not isinstance(value, np.ndarray):
            raise ValueError("Parameter not expected")

        if value.dtype == np.int:
            value_idx = self.data["EntryNr2"].searchsorted(value)
            if not (self.data["EntryNr2"][value_idx] == value).all():
                missing_matches = value[self.data["EntryNr2"][value_idx] != value]
                raise ValueError(
                    u"Not all entries found in Matches: {0!r:s}".format(missing_matches)
                )
        elif value.dtype == self.data.dtype:
            raise NotImplementedError(u"setitem with replacement view not implemented")
        else:
            raise ValueError(
                u"Unexpected dtype in parameter: {0!:s}".format(value.dtype)
            )
        self._set_relationflags_on_rows(value_idx)

    def _set_relationflags_on_rows(self, row_indexes):
        """method which sets the column flags to the appropriate
        relation type for the indexes of the data buffer. The base
        class does not change any flags. """
        pass


class CanonicalBestMatchesOfEntry(RelationsOfEntry):
    """"variant that only looks at the main splicing variant"""

    @staticmethod
    def filter(row):
        return row["isCanonicalSplicing"]

    def _set_relationflags_on_rows(self, row_indexes):
        mask = np.zeros(len(self.data), dtype="bool")
        mask[row_indexes] = True
        self.data["isCanonicalSplicing"] = mask


class StablePairsOfEntry(RelationsOfEntry):
    """Stable Pairs have to be Canonical Splicing and set isSP to true"""

    @staticmethod
    def filter(row):
        return np.logical_and(row["isSP"], row["isCanonicalSplicing"])

    def _set_relationflags_on_rows(self, row_indexes):
        mask = np.zeros(len(self.data), dtype="bool")
        mask[row_indexes] = True
        self.data["isSP"] = mask
        self.data["isCanonicalSplicing"][row_indexes] = True


class VPairsOfEntry(RelationsOfEntry):
    """Verifiy Pairs have to be Canonical Splicing and set isVP to true.
    So far we do not enforce isSP to be true as well, because of consistency
    step where we potentially augment non-sps to vps."""

    @staticmethod
    def filter(row):
        return np.logical_and(row["isVP"], row["isCanonicalSplicing"])

    def _set_relationflags_on_rows(self, row_indexes):
        mask = np.zeros(len(self.data), dtype="bool")
        mask[row_indexes] = True
        self.data["isVP"] = mask
        self.data["isCanonicalSplicing"][row_indexes] = True


class RelationManager(object):
    def __init__(self, data):
        self.data = data
        self._data_changed = False

    def __getitem__(self, key):
        if key == "SP":
            res = StablePairsOfEntry(self.data)
        elif key == "VP":
            res = VPairsOfEntry(self.data)
        elif key == "BM":
            res = CanonicalBestMatchesOfEntry(self.data)
        elif key == "ALL":
            res = RelationsOfEntry(self.data)
        else:
            raise KeyError(u"Unexpected key: {0!r:s}".format(key))
        return res

    def __setitem__(self, key, value):
        rels_obj = self.__getitem__(key)
        rels_obj.set_relations(value)
        self._data_changed = True

    def is_data_insync(self):
        return not self._data_changed


class GenomePair(object):
    def __init__(self, mtab, itab):
        self.matches = mtab
        self.entry_offset = itab
        self.rels = [None] * len(itab)

    def __getitem__(self, item):
        if not isinstance(item, int):
            raise IndexError(u"Invalid index or slice: {0!r:s}".format(item))
        if item < 0:
            raise IndexError(u"Invalid index or slice: {0!r:s}".format(item))
        if item < 0:
            raise IndexError(u"Invalid index or slice: {0!r:s}".format(item))
        if item < 0:
            item += len(self.entry_offset)
        else:
            item -= 1
        if item >= len(self.entry_offset):
            raise IndexError(u"Index is out of bound: {}".format(item))
        if not self.rels[item] is None:
            return self.rels[item]

        range_of_entry = self.entry_offset[item]
        data = self.matches[range_of_entry[0] : range_of_entry[1]]
        self.rels[item] = RelationManager(data)
        return self.rels[item]

    def flush(self):
        modified = 0
        for i, rel_man in enumerate(self.rels):
            if rel_man is not None and not rel_man.is_data_insync():
                modified += self.matches.modify_rows(
                    start=self.entry_offset[i, 0],
                    stop=self.entry_offset[i, 1],
                    rows=rel_man.data,
                )
                rel_man._data_changed = False

    def close(self):
        self.flush()
        del self.rels


class OmaDB(object):
    def __init__(self, file, mode="r"):
        if mode != "r":
            filters = tables.Filters(complevel=6, complib="zlib")
        else:
            filters = None
        self.db_handle = tables.open_file(file, mode=mode, filters=filters)
        self.genome_data = self.db_handle.get_node("/GenomeSummary")

    @lru_cache(maxsize=256)
    def matches(self, genome1, genome2):
        try:
            mtab = self.db_handle.get_node(
                "/Matches/{}/{}/Relations".format(genome1, genome2)
            )
            itab = self.db_handle.get_node(
                "/Matches/{}/{}/ProteinIndex".format(genome1, genome2)
            )
        except tables.NoSuchNodeError:
            raise KeyError(u"genome pair does not exist in database")
        gs = self.genome_data.read_where("UniProtSpeciesCode==b'{}'".format(genome1))
        if len(itab) != gs["TotEntries"]:
            raise OmaDBError(
                u"Nr of Protein does not match: {} vs {}".format(
                    len(itab), gs["TotEntries"]
                )
            )
        return GenomePair(mtab, itab)

    def close(self):
        self.db_handle.close()


class OmaDBError(Exception):
    pass



import random
import unittest
import os
import Bio.Seq
import Bio.Data.CodonTable
import pyoma.browser.db as pyomadb
import tables
import numpy


class DatabaseChecks(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        try:
            path = os.environ["PYOMA_DB2CHECK"]
        except KeyError:
            raise unittest.SkipTest("No database specified in PYOMA_DB2CHECK")

        cls.db = pyomadb.Database(path)

    def translated_cdna_match_protein_sequence(self, cdna, prot):
        cdna = cdna.replace("X", "N")
        for tab in Bio.Data.CodonTable.generic_by_id.keys():
            tab_ok = True
            trans = Bio.Seq.translate(cdna, table=tab)
            if not 3 >= len(trans) - len(prot) >= 0:
                return False
            for pos, (trans_aa, prot_aa) in enumerate(zip(trans, prot)):
                if trans_aa == prot_aa or trans_aa == "X" or prot_aa == "X":
                    continue
                elif prot_aa == "M" and pos == 0 and trans_aa != "*":
                    continue
                else:
                    tab_ok = False
                    break
            if tab_ok:
                return True

    def test_cdna_and_protein_sequence_match(self):
        """test translated cdna sequence and protein sequence match.

        This is done for a random sample of 1000 entries"""
        SAMPLES = 1000
        nr_entries = self.db.id_resolver.max_entry_nr
        for entry_nr in random.sample(range(nr_entries + 1), SAMPLES):
            with self.subTest(entry_nr=entry_nr):
                cdna = self.db.get_cdna(entry_nr).decode()
                prot = self.db.get_sequence(entry_nr).decode()
                self.assertTrue(self.translated_cdna_match_protein_sequence(cdna, prot))

    def test_increasing_offsets(self):
        entry_tab = self.db.get_hdf5_handle().get_node("/Protein/Entries")
        seq_off = -1
        cds_off = -1
        for row in entry_tab:
            self.assertLess(
                seq_off,
                row["SeqBufferOffset"],
                "SeqBufferOffset decreases in row {}: {} vs {}".format(
                    row.nrow, seq_off, row["SeqBufferOffset"]
                ),
            )
            self.assertLess(
                cds_off,
                row["CDNABufferOffset"],
                "CDNABufferOffset decreases in row {}: {} vs {}".format(
                    row.nrow, seq_off, row["CDNABufferOffset"]
                ),
            )
            seq_off = row["SeqBufferOffset"]
            cds_off = row["CDNABufferOffset"]

    def test_homeology_flag(self):
        genome_tab = self.db.get_hdf5_handle().get_node("/Genome")
        for g in (b"WHEAT", b"GOSHI", b"BRANA"):
            for row in genome_tab.read_where("UniProtSpeciesCode == g"):
                self.assertTrue(
                    row["IsPolyploid"],
                    "{} is not recorded as polyploid genome".format(g),
                )
        for g in (b"YEAST", b"HUMAN", b"PLAF7", b"ARATH", b"MOUSE"):
            for row in genome_tab.read_where("UniProtSpeciesCode == g"):
                self.assertFalse(
                    row["IsPolyploid"],
                    "{} is recorded to be a ployploid genome".format(g),
                )

    def test_synteny_scores_exist(self):
        for g in ("WHEAT", "BRANA", "GOSHI"):
            try:
                t = self.db.get_hdf5_handle().get_node(
                    "/PairwiseRelation/{}/within".format(g)
                )
            except tables.NoSuchNodeError:
                # if species does not exist, we skip - not all datasets will have these genomes
                continue
            syn_col = t.col("SyntenyConservationLocal")
            computed_pairs = numpy.where(syn_col >= 0)
            self.assertLess(
                0, len(computed_pairs[0]), "No synteny values computed for {}".format(g)
            )

    def test_alignment_of_pairwise_orthologs(self):
        SAMPLES = 500
        nr_entries = self.db.id_resolver.max_entry_nr
        seq_search = self.db.seq_search
        for entry_nr1 in random.sample(range(1, nr_entries + 1), 10 * SAMPLES):
            vps = self.db.get_vpairs(entry_nr1)
            if len(vps) == 0:
                continue
            pair = random.choice(vps)
            with self.subTest(pair=pair):
                s1 = self.db.get_sequence(pair["EntryNr1"])
                matches = [(pair["EntryNr2"],)]
                alignment = list(
                    seq_search._align_entries(s1, matches, compute_distance=True)
                )[0]
                score = alignment[1][0]
                dist = alignment[1][2]
                if pair["Score"] < 1:
                    self.assertGreaterEqual(score, 120)
                else:
                    self.assertAlmostEqual(pair["Score"], score, places=1)
                    self.assertAlmostEqual(pair["Distance"], dist, places=1)
            SAMPLES -= 1
            if SAMPLES <= 0:
                break

    def test_cached_orthologs_count(self):
        SAMPLES = 80
        nr_entries = self.db.id_resolver.max_entry_nr
        for entry_nr in random.sample(range(nr_entries + 1), SAMPLES):
            with self.subTest(entry_nr=entry_nr):
                cached_ortholog_cnts = self.db.nr_ortholog_relations(entry_nr)
                vps = self.db.get_vpairs(entry_nr)
                self.assertEqual(len(vps), cached_ortholog_cnts["NrPairwiseOrthologs"])

                entry = self.db.entry_by_entry_nr(entry_nr)
                if entry["OmaGroup"] > 0:
                    oma_grp_memb = self.db.oma_group_members(entry["OmaGroup"])
                    self.assertEqual(
                        len(oma_grp_memb) - 1,
                        cached_ortholog_cnts["NrOMAGroupOrthologs"],
                    )

                if entry["OmaHOG"] != b"":
                    hog_ind_orth = self.db.get_hog_induced_pairwise_orthologs(entry)
                    self.assertEqual(
                        len(hog_ind_orth),
                        cached_ortholog_cnts["NrHogInducedPWOrthologs"],
                    )
